import torch
from hybra.utils import kappa_alias

class HybrALoss(torch.nn.Module):
    def __init__(self, base_loss:torch.nn.Module,decimation_factor: int, beta: float = 0.0, gamma: float = 0.0):
        super().__init__()

        self.base_loss = base_loss
        self.decimation_factor = decimation_factor
        self.beta = beta
        self.gamma = gamma

    def forward(self, prediction, target, filterbank):
        loss = self.base_loss(prediction, target)
        kappa, alias = kappa_alias(filterbank._filters.squeeze(1), self.decimation_factor)

        return loss, loss + self.beta*(kappa-1) + self.gamma * torch.sum(alias)


import torch
import torch.nn as nn
import torch.nn.functional as F
from hybra.utils import fir_tightener3000, random_filterbank, kappa_alias

class HybrA(nn.Module):
    def __init__(self, path_to_auditory_filter_config, start_tight=True):
        super().__init__()
        
        config = torch.load(path_to_auditory_filter_config, weights_only=False, map_location="cpu")

        self.signal_length = 0
        self.audlet_real = config['auditory_filters_real'].clone().detach()
        self.audlet_imag = config['auditory_filters_imag'].clone().detach()
        self.audlet = self.audlet_real.squeeze(1) + 1j * self.audlet_imag.squeeze(1)
        self.audlet_stride = config['auditory_filters_stride']
        self.audlet_length = self.audlet_real.shape[-1]
        self.audlet_channels = config['n_filters']
        self.kernel_length = config['kernel_size']
        
        random_kernels = random_filterbank(N=self.audlet_length, J=1, T=self.kernel_length, norm=True, support_only=False)
        
        if start_tight:
#             random_kernels = fir_tightener4000(random_kernels.squeeze(1), self.encoder_length, 1,eps=1.1)
            random_kernels = fir_tightener3000(random_kernels, self.kernel_length, D=1, eps=1.01)
            random_kernels = torch.cat(self.audlet_channels * [random_kernels], dim=0)          

        self.kernels = nn.Parameter(random_kernels, requires_grad=True)
        self.filters = torch.empty(self.audlet.shape)
        self._filters = torch.empty(self.audlet.shape) # this is the one to optimize

    def forward(self, x:torch.Tensor) -> torch.Tensor:
        """Forward pass of the HybridFilterbank.

        Parameters:
        -----------
        x (torch.Tensor) - input tensor of shape (batch_size, 1, signal_length)

        Returns:
        --------
        x (torch.Tensor) - output tensor of shape (batch_size, num_channels, signal_length//hop_length)
        """
        self.signal_length = x.shape[-1]
        x = F.pad(x, (0, x.shape[-1] % self.audlet_stride), mode='constant', value=0)
        hybra_kernels = torch.fft.ifft(
            torch.fft.fft(self.audlet.to(x.device).squeeze(1), dim=1) *
            torch.fft.fft(self.kernels.squeeze(1), dim=1),
            dim=1
            ).unsqueeze(1)
    
        self._filters = hybra_kernels
        self.filters = hybra_kernels.clone().detach()
        padding_length = self.filters.shape[-1] - 1

        output_real = F.conv1d(
            F.pad(x.unsqueeze(1), (padding_length, 0), mode='circular'),
            torch.fliplr(hybra_kernels.real),
            stride=self.audlet_stride,
        )

        output_imag = F.conv1d(
            F.pad(x.unsqueeze(1), (padding_length, 0), mode='circular'),
            torch.fliplr(hybra_kernels.imag),
            stride=self.audlet_stride,
        )

        return output_real + 1j * output_imag

    def encoder(self, x:torch.Tensor):
        """For learning use forward method

        """
        padding_length = self.filters.shape[-1] - 1

        return F.conv1d(
            F.pad(x.unsqueeze(1), (padding_length, 0), mode='circular'),
            torch.fliplr(self.filters.real.to(x.device)),
            stride = self.audlet_stride,
        ) + 1j * F.conv1d(
            F.pad(x.unsqueeze(1), (padding_length, 0), mode='circular'),
            torch.fliplr(self.filters.imag.to(x.device)),
            stride = self.audlet_stride,
        )

    def decoder(self, x:torch.Tensor) -> torch.Tensor:
        """Forward pass of the dual HybridFilterbank.

        Parameters:
        -----------
        x (torch.Tensor) - input tensor of shape (batch_size, n_filters, signal_length//hop_length)

        Returns:
        --------
        x (torch.Tensor) - output tensor of shape (batch_size, signal_length)
        """

        padded_signal_length = x.shape[-1] * self.audlet_stride
        padding_length = self.filters.shape[-1] - 1
        Ls = x.shape[-1] + padding_length
        kernel_size = self.filters.real.shape[-1]
        output_padding_length = int((padded_signal_length - kernel_size - (Ls-1) * self.audlet_stride) / -2) 
        
        x_real = x.real
        x_imag = x.imag

        x = (
            F.conv_transpose1d(
                F.pad(x_real, (0, padding_length), mode='circular'),
                torch.fliplr(self.filters.real),
                stride=self.audlet_stride,
                padding=output_padding_length
            )
            + F.conv_transpose1d(
                F.pad(x_imag, (0, padding_length), mode='circular'),
                torch.fliplr(self.filters.imag),
                stride=self.audlet_stride,
                padding=output_padding_length
            )
        )
        
        return torch.roll(2*self.audlet_stride * x.squeeze(1), output_padding_length - (kernel_size-1))[:, :self.signal_length]

    @property
    def condition_number(self):
        return float(kappa_alias(self.filters.squeeze(1), self.audlet_stride, aliasing=False))



import torch

def random_filterbank(N:int, J:int, T:int, norm:bool=True, support_only:bool=False) -> torch.Tensor:
    """
    Constructs a [J, N] tensor with J i.i.d. Gaussian random filters of length N with support T.
    Input:  N: Signal length
            J: Number of filters
            T: Filter support (or length of conv1d kernels). Default: T=N
            norm: If True then the the variance is 1/(TJ)
            support_only: If True then N <- T 
    Output: Impulse responses of the random filterbank (torch.tensor[J, T])
    """
    if T == None:
        T = N
    if norm:
        w = torch.randn(J, T).div(torch.sqrt(torch.tensor(J*T)))
    else:
        w = torch.randn(J, T)
    if support_only:
        w_cat = w
    else:
        w_cat = torch.cat([w, torch.zeros(J, N-T)], dim=1)
    return w_cat
    
def kappa_alias(w:torch.Tensor, D:int, aliasing:bool=True) -> torch.Tensor:
    """
    Computes the condition number and the norm of the aliasing term of a filterbank using the polyphase representation.
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[J, T]
            D: Decimation (or downsampling) factor, must divide filter length!
            aliasing: If False, only the condition umber is returned
    Output: Condition number and norm of the aliasing term
    """
    w_hat = torch.fft.fft(w, dim=-1).T
    kappa = condition_number(w_hat, D)

    if aliasing:
        alias = torch.zeros_like(w_hat)
        N = alias.shape[0]
        for j in range(1,D):
            alias += w_hat * torch.conj(w_hat.roll(j * N//D, 0))
        alias = torch.sum(alias, dim=1)
        return kappa, torch.linalg.norm(alias)
    else:
        return kappa

def condition_number(w_hat:torch.Tensor, D:int) -> torch.Tensor:
    """
    Computes the condition number of a filterbank w_hat using the polyphase representation.
    Input:  w: Frequency responses of the filterbank as 2-D Tensor torch.tensor[length, num_channels]
            D: Decimation (or downsampling) factor, must divide filter length!
    Output: Condition number.
    """
    if D == 1:
        lp = torch.sum(w_hat.abs() ** 2, dim=1)
        A = torch.min(lp)
        B = torch.max(lp)
        return B/A
    else:    
        N = w_hat.shape[0]
        J = w_hat.shape[1]
        assert N % D == 0, "Oh no! Decimation factor must divide signal length!"

        A = torch.tensor([torch.inf]).to(w_hat.device)
        B = torch.tensor([0]).to(w_hat.device)
        Ha = torch.zeros((D,J)).to(w_hat.device)
        Hb = torch.zeros((D,J)).to(w_hat.device)

        for j in range(N//D):
            idx_a = (j - torch.arange(D) * (N//D)) % N
            idx_b = (torch.arange(D) * (N//D) - j) % N
            Ha = w_hat[idx_a, :]
            Hb = torch.conj(w_hat[idx_b, :])
            lam = torch.linalg.eigvalsh(Ha @ Ha.H + Hb @ Hb.H).real
            A = torch.min(A, torch.min(lam))
            B = torch.max(B, torch.max(lam))
        return B/A

def can_tight(w:torch.Tensor, D:int) -> torch.Tensor:
    """
    Computes the canonical tight filterbank of w (time domain) using the polyphase representation.
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[num_channels, length]
            D: Decimation (or downsampling) factor, must divide filter length!
    Output: Canonical tight filterbank of W (torch.tensor[num_channels, length])
    """
    w_hat = torch.fft.fft(w.T, dim=0)
    if D == 1:
        lp = torch.sum(w_hat.abs() ** 2, dim=1).reshape(-1,1)
        w_hat_tight = w_hat * (lp ** (-0.5))
        return torch.fft.ifft(w_hat_tight.T, dim=1)
    else:
        N = w_hat.shape[0]
        J = w_hat.shape[1]
        assert N % D == 0, "Oh no! Decimation factor must divide signal length!"

        w_hat_tight = torch.zeros(J, N, dtype=torch.complex64)
        for j in range(N//D):
            idx = (j - torch.arange(D) * (N//D)) % N
            H = w_hat[idx, :]
            U, _, V = torch.linalg.svd(H, full_matrices=False)
            H = U @ V
            w_hat_tight[:,idx] = H.T.to(torch.complex64)
        return torch.fft.ifft(torch.fft.ifft(w_hat_tight.T, dim=1) * D ** 0.5, dim=0).T

def fir_tightener3000(w, supp, D, eps=1.01, Ls=None):
    """
    Iterative tightening procedure with fixed support for a given filterbank w
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[num_channels, length].
            supp: Desired support of the resulting filterbank
            D: Decimation (or downsampling) factor, must divide filter length!
            eps: Desired condition number
            Ls: control syste length
    Output: Filterbank with condition number *kappa* and support length *supp*. If length=supp then the resulting filterbank is the canonical tight filterbank of w.
    """
    if Ls is not None:
        w =  torch.cat([w, torch.zeros(w.shape[0], Ls-w.shape[1])], dim=1)
    w_tight = w.clone()
    kappa = kappa_alias(w, D, aliasing=False)
    while kappa > eps:
        w_tight = can_tight(w_tight, D)
        w_tight[:, supp:] = 0
        kappa = kappa_alias(w_tight, D, aliasing=False)
    if Ls is None:
        return w_tight
    else:
        return w_tight[:,:supp]

def fir_tightener4000(w, supp, eps=1.01):
    """
    Iterative tightening procedure with fixed support for a given filterbank w. Every filter will form an (approximate) tight frame.
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[num_channels, length].
            supp: Desired support of the resulting filterbank
            eps: Desired condition number
    Output: Filterbank with support length *supp* and every filter has condition number *kappa*.
    """
    D=1 # only works without decimation!
    for i in range(w.shape[0]):
        filter = w[i,:].reshape(1,-1)
        w[i,:] = fir_tightener3000(filter, supp, D, eps)
    return w

def smooth_fir(w_hat:torch.Tensor, supp:int, time_domain:bool=False) -> torch.Tensor:
    """
    Takes a filterbank in frequency domain (as columns) and constructs a smoothed FIR version with support length *support*.
    Input:  w: Frequency responses of the filterbank as 2-D Tensor torch.tensor[length, num_channels]
            supp: Desired support in time domain
            time_domain: If True, w is treated as containing impulse responses torch.tensor[num_channels, length]
    Output: Impulse responses
    """
    g = torch.exp(-torch.pi * torch.arange(-supp//2,supp//2)**2 / ((supp-12)/2)**2)
    g = g.reshape(1,-1)
    if time_domain:
        w = w_hat
    else:
        w = torch.fft.ifft(w_hat, dim=0).T
        w = w.roll(supp//2, 0)[:, :supp]
    return w * g